#include <cstdio>
#include <cstring>
#include <iconv.h>
#include <unistd.h>
#include <pqxx/pqxx>
#include <ft2build.h>
#include FT_FREETYPE_H

#include "glwindow.h"

iconv_t ucs4_iconv;

// UCS-4 string
class widestring : public std::basic_string<unsigned>
{
public:
	void operator= (const char *from)
	{
		unsigned bytes = std::strlen(from);
		char *from_buf = strdup(from);
		unsigned *to_buf = new unsigned[bytes + 1];

		char *inptr = from_buf, *outptr = reinterpret_cast<char *> (to_buf);

		size_t in_left = bytes;
		size_t out_left = bytes * sizeof(unsigned);

		size_t ret = iconv(ucs4_iconv, NULL, NULL, &outptr, &out_left);
		if (ret == (size_t)(-1)) {
			throw std::runtime_error("Error in iconv during initialization");
		}

		ret = iconv(ucs4_iconv, &inptr, &in_left, &outptr, &out_left);
		if (ret == (size_t)(-1)) {
			perror("iconv");
			throw std::runtime_error("Error in iconv during conversion");
		}

		erase(begin(), end());
		std::copy(to_buf, reinterpret_cast<unsigned *> (outptr), std::back_inserter(*this));

		free(from_buf);
		delete[] to_buf;
	}
};

template<>
void std::char_traits<unsigned>::assign(unsigned &to, unsigned const &from)
{
	to = from;
}

template<>
unsigned *std::char_traits<unsigned>::copy(unsigned *to, unsigned const *from, unsigned n)
{
	return static_cast<unsigned *>(memcpy(to, from, n * sizeof(unsigned)));
}

template<>
unsigned *std::char_traits<unsigned>::move(unsigned *to, unsigned const *from, unsigned n)
{
	return static_cast<unsigned *>(memmove(to, from, n * sizeof(unsigned)));
}

template<>
unsigned *std::char_traits<unsigned>::assign(unsigned *to, size_t n, unsigned a)
{
	for (unsigned i = 0; i < n; ++i)
		*to++ = a;
	return to;
}


template<>
void pqxx::from_string<widestring>(const char *from, widestring &to)
{
	to = from;
}

class Tournament {
public:
	int id;
	widestring name;
};

Tournament active_tournament;
FT_Face font;

/* A trigger that sets a flag whenever it's trigged. */
class FlagTrigger : pqxx::trigger {
private:
	bool flag;
	
public:
	FlagTrigger(pqxx::connection_base &conn, const PGSTD::string &name)
		: pqxx::trigger(conn, name), flag(false) {}
	virtual ~FlagTrigger() throw () {}
	
	virtual void operator() (int pid)
	{
		flag = true;
		std::fprintf(stderr, "Received a flag trigger from pid %u\n", pid);
	}

	bool get_flag() const
	{
		return flag;
	}

	void reset_flag()
	{
		flag = false;
	}
};

/* A transactor that fetches the current tournament and some information about it. */
class FetchCurrentTournament : public pqxx::transactor<> {
private:
	Tournament *tourn;

public:
	FetchCurrentTournament(Tournament *tourn) : tourn(tourn) {}
	void operator() (pqxx::transaction<> &t)
	{
		pqxx::result res( t.exec("SELECT * FROM bigscreen.active_tournament NATURAL JOIN tournaments") );
		try {
			pqxx::result::tuple tournament = res.at(0);

			tourn->id = tournament["tournament"].as(tourn->id);
			tourn->name = tournament["tournamentname"].as(tourn->name);
		} catch (PGSTD::out_of_range &e) {
			tourn->id = -1;
			tourn->name = "";
		}
	}
};

void init(pqxx::connection &conn)
{
	conn.perform(FetchCurrentTournament(&active_tournament));

	if (active_tournament.id == -1) {
		std::fprintf(stderr, "No active tournament\n");
	} else {
		std::fprintf(stderr, "Current tournament is %d (name: '%s')\n",
			active_tournament.id, active_tournament.name.c_str());
	}
}

void main_loop(pqxx::connection &conn)
{
	if (active_tournament.id == -1) {
		// No active tournament, sleep a second or so and exit
		sleep(1);
		return;
	}
	
	pqxx::work t(conn, "trx");

	// fetch all songs
	pqxx::result res( t.exec("SELECT * FROM songs") );
	for (pqxx::result::const_iterator i = res.begin(); i != res.end(); ++i) {
		std::fprintf(stderr, "%s\n", i["title"].c_str());
	}
	t.commit();
	
	sleep(1);
}

void init_freetype()
{
	FT_Library library;
	if (FT_Init_FreeType(&library))
		throw std::logic_error("FreeType init failed.");
	if (FT_New_Face(library, "/usr/share/fonts/truetype/msttcorefonts/Georgia.ttf", 0, &font))
		throw std::logic_error("Face opening failed.");
	if (FT_Set_Char_Size(font, 0, 12, 96, 96))
		throw std::logic_error("Size set failed.");
}

int main(int argc, char **argv)
{
	ucs4_iconv = iconv_open("ucs-4", "utf-8");
	
	GLWindow glw("CCBS bigscreen", 800, 600, 32, false, 16, -1);
	
	try {
		init_freetype();
		pqxx::connection conn("dbname=ccbs host=altersex.samfundet.no user=ccbs password=GeT|>>B_");
		FlagTrigger tournament_changed(conn, "active_tournament");
		
		// when active_tournament is changed, we destroy everything and start from scratch
		for ( ;; ) {
			tournament_changed.reset_flag();
			init(conn);
			do {
				main_loop(conn);
				conn.get_notifs();
			} while (!tournament_changed.get_flag());
			std::fprintf(stderr, "active_tournament changed, resetting...\n");
		}
	} catch (const std::exception &e) {
		std::fprintf(stderr, "Exception: %s\n", e.what());
		exit(1);
	}
	
	return 0;
}
